#!/usr/bin/perl -w

# admprofs.pl --- Display tabulated results for various tests that
#   the college uses to compare students by score category and how
#   far along in the admissions process they are.

# Author: Michael Olson
# Contact by email: mwolson at gnu.org
# Usage: admprofs.pl sess yr score_source

# License: The GNU General Public License (GPL).  What this means is
#   that there are no restrictions on the use of this program.  To
#   distribute it or a derivative work, however, the distributors must
#   make their version of this particular program available under the
#   terms of the GPL.  For more information, please consult the
#   following web page: [http://www.gnu.org/licenses/gpl.html].

# Warranty: The author explicitly refuses to warrant or make any sort
#   of guarantee that this program will work a particular way.  Use at
#   your own risk.

# The benefits of using Perl instead of making a report are:
#
# - It will be much easier to maintain since this report handles all
#   of the act/sat/rank variants in one program.
#
# - Modularity may be more easily achieved.  In other words, it
#   shouldn't be difficult at all to add a new type of profile report
#   to this.
#
# - You have much *much* more control over how the data is processed.
#
# - The output from this program may be easily read by a web browser.

# Instructions for adding a new test or score source:
#
# We'll use the shorthand `{new_source}' to mean the name of your new
# score source.
#
# - In the `Global variables'->`Scoring categories' section, add a
#   line for %{new_source}_cat in %na_cat to define what you want to
#   use for the "N/A" category, like:
#
#       "{new_source}" => "No score",
#
# - In the `Global variables'->`Scoring categories' section, add a new
#   entry to %score_choices, like:
#
#       "{new_source}" => "Description of new source",
#
# - Define the new query in the `Define SQL queries' section.  Make
#   sure it assigns an integer value to `score' so that nothing can
#   fall between the gaps in the categories.  Make sure that those
#   with invalid scores are not included in the results of the query.
#
# - If your query needs to represent decimal numbers, convert them to
#   integers in the query by multiplying by 10^N, where N is the
#   number of decimal places.  Round the result.  Then add a line like
#   the following to %score_divisor in the `Global
#   variables'->`Scoring categories' section:
#
#       "{new_source}" => 100,
#
#   This would cause the average score to be divided by 100 before
#   being displayed.
#
# - Define the way categories are interpreted in the `Define score
#   categories' section.  Order matters; the first category entered
#   will come first in the output, and so on.
#
# - Make sure that every result returned from ${new_source}_q can be
#   classified by exactly one category.  Values should not fall into
#   more than one category, since the categories are not checked in
#   any particular order.  If a value is returned from the score query
#   that does not match a category, it could possibly throw off the
#   total count.

# We care about making clear code
use strict;

# Load the database interface module
use DBI;
use DBI qw(:sql_types);

### Global variables

# Queries
my $stu_q;               # get qualified students
my $stu_progress_q;      # students and their admissions progress code
my %score_q;             # a hash of score type and the query we will
                         # use to get a student's score

# Scoring categories
my %score_cat;             # hash of score types to an ordered hash of
                           # score categories and variables that
                           # interpret them
my %na_cat =               # what to call the scoring category that
                           # contains all of those who didn't match any
                           # categories
    (
     "act"  => "N/A",
     "gpa"  => "N/A",
     "rank" => "N/A",
     "satm" => "N/A",
     "satv" => "N/A",
    );

my %score_choices =             # available scores and descriptions
    (
     "act"  => "ACT Score",
     "gpa"  => "GPA",
     "rank" => "Class Rank",
     "satm" => "SAT Math Score",
     "satv" => "SAT Verbal Score",
    );

my %score_divisor =             # what to divide score by to obtain
                                # averages
    (
     "gpa"  => 100,
     "rank" => 1,
    );

# Parameters
my $prog_name = $0;             # name of our program
my $score_type;                 # type of score to find
my $sess;                       # session, like "FA"
my $yr;                         # year

# Programmtically determined data
my $dbh;                     # handle for database
my %percentiles;             # 25th and 75th percentiles
my @progresses =             # ordered list of progress codes
    ("Applied", "Accepted", "Deposit Paid");
my %stu_progress;            # students and admissions progress
my %stu_score;               # students and score
my %progress_score_cnt;      # count of students by progress and score
my %progress_summary;        # count and average score of each progress

# Report customizations
my $organization = "Bethel College";
my $student_type = "Freshman";

### Define SQL queries

# Define which students we are interested in.
$stu_q = q{
select  distinct adm_rec.id,
        case when ctc_rec.resrc in ("ENROLDEP", "ACCEPTED", "TEMPACCE")
          then ctc_rec.resrc
        else "NA"
        end resrc,
        ctc_rec.stat
from    adm_rec,
        ctc_rec
where   adm_rec.plan_enr_sess = ?
 and    adm_rec.plan_enr_yr = ?
 and    adm_rec.id = ctc_rec.id
 and    adm_rec.subprog = "TRAD"
 and    adm_rec.cl in ("FF", "FR")
 and    adm_rec.enrstat in ("APPLIED", "ACCEPTED", "TEMPACCE",
         "READMIT", "REGCOMPL", "COMPLAPP", "DENIED",
         "CANCEL", "REACTIVA")
into temp t_stu with no log;
};                              # end of query

# Split students by their admissions progress; duplicate progresses
# are allowed because if a student is Accepted, they must have already
# applied.
$stu_progress_q =               # Applied category
    q{
select  distinct t_stu.id,
        "Applied" progress
from    t_stu

union
} .                             # Accepted category
    q{
select  distinct t_stu.id,
        "Accepted" progress
from    t_stu
where   t_stu.stat = "C"
 and    t_stu.resrc in ("ACCEPTED", "TEMPACCE")

union
} .                             # Deposit Paid category
    q{
select  distinct t_stu.id,
        "Deposit Paid" progress
from    t_stu,
        t_stu t1
where   t_stu.id = t1.id
 and    t_stu.stat = "C"
 and    t_stu.resrc in ("ACCEPTED", "TEMPACCE")
 and    t1.stat = "C"
 and    t1.resrc = "ENROLDEP"
into temp t_progress with no log;
};                              # end of query

# Get ACT score of each student
$score_q{act} = q{
select  distinct t_stu.id,
        exam_rec.score5 score
from    t_stu,
        exam_rec
where   t_stu.id = exam_rec.id
 and    exam_rec.ctgry = "HACT"
 and    exam_rec.score5 > 0
 and    exam_rec.score5 <= 36
into temp t_score with no log;
};                              # end of query

# Get SAT math score of each student
$score_q{satm} = q{
select	distinct t_stu.id,
        exam_rec.score1 score
from    t_stu,
        exam_rec
where   t_stu.id = exam_rec.id
 and    exam_rec.ctgry = "HSAT"
 and    exam_rec.score1 > 0
 and    exam_rec.score1 <= 800
into temp t_score with no log;
};                              # end of query

# Get SAT verbal score of each student
$score_q{satv} = q{
select	distinct t_stu.id,
        exam_rec.score2 score
from    t_stu,
        exam_rec
where   t_stu.id = exam_rec.id
 and    exam_rec.ctgry = "HSAT"
 and    exam_rec.score2 > 0
 and    exam_rec.score2 <= 800
into temp t_score with no log;
};                              # end of query

# Get high school class rank percentile of each student
$score_q{rank} = q{
select	distinct t_stu.id,
        case when 100 * ed_rec.rank / ed_rec.cl_size < 1
          then 1
        else
          round (100 * ed_rec.rank / ed_rec.cl_size)
        end score
from	t_stu,
        ed_rec,
        sch_rec
where   t_stu.id = ed_rec.id
 and    ed_rec.sch_id = sch_rec.id
 and    sch_rec.ctgry = "HS"
 and    ed_rec.cl_size > 20
 and    ed_rec.rank <= ed_rec.cl_size
 and    ed_rec.rank > 0
into temp t_score with no log;
};                              # end of query

# Get high school GPA of each student
$score_q{gpa} = q{
select	distinct t_stu.id,
        case when ed_rec.gpa > 4
          then 400
        else
          round (ed_rec.gpa * 100)
        end score
from	t_stu,
        ed_rec,
        sch_rec
where   t_stu.id = ed_rec.id
 and    ed_rec.sch_id = sch_rec.id
 and    sch_rec.ctgry = "HS"
 and    ed_rec.gpa > 0
into temp t_score with no log;
};                              # end of query

### Define scoring categories
#
# Format:
# "description" => { lower => lower_bound, upper => upper_bound,
#                    order => display_order };
#
# When data is counted, it will only use the first matching category.

# ACT scoring categories
$score_cat{act} = {
    "33-36" => { lower => 33, upper => 36, order => 1 },
    "30-32" => { lower => 30, upper => 32, order => 2 },
    "27-29" => { lower => 27, upper => 29, order => 3 },
    "24-26" => { lower => 24, upper => 26, order => 4 },
    "21-23" => { lower => 21, upper => 23, order => 5 },
    "18-20" => { lower => 18, upper => 20, order => 6 },
    "15-17" => { lower => 15, upper => 17, order => 7 },
    "12-14" => { lower => 12, upper => 14, order => 8 },
    "9-11"  => { lower =>  9, upper => 11, order => 9 },
    "6-8"   => { lower =>  6, upper =>  8, order => 10 },
    "1-5"   => { lower =>  1, upper =>  5, order => 11 },
};

# SAT math scoring categories
$score_cat{satm} = {
    "750-800" => { lower => 750, upper => 800, order => 1 },
    "700-749" => { lower => 700, upper => 749, order => 2 },
    "650-699" => { lower => 650, upper => 699, order => 3 },
    "600-649" => { lower => 600, upper => 649, order => 4 },
    "550-599" => { lower => 550, upper => 599, order => 5 },
    "500-549" => { lower => 500, upper => 549, order => 6 },
    "450-499" => { lower => 450, upper => 499, order => 7 },
    "400-449" => { lower => 400, upper => 449, order => 8 },
    "350-399" => { lower => 350, upper => 399, order => 9 },
    "300-349" => { lower => 300, upper => 349, order => 10 },
    "250-299" => { lower => 250, upper => 299, order => 11 },
    "200-249" => { lower => 200, upper => 249, order => 12 },
};

# SAT verbal scoring categories
$score_cat{satv} = $score_cat{satm};

# Class rank categories, in deciles (smaller decile numerically means
# higher rank, lower score means higher rank)
$score_cat{rank} = {
    "1" => { lower => 1, upper => 10, order => 1 },
    "2" => { lower => 11, upper =>  20, order => 2 },
    "3" => { lower => 21, upper =>  30, order => 3 },
    "4" => { lower => 31, upper =>  40, order => 4 },
    "5" => { lower => 41, upper =>  50, order => 5 },
    "6" => { lower => 51, upper =>  60, order => 6 },
    "7" => { lower => 61, upper =>  70, order => 7 },
    "8" => { lower => 71, upper =>  80, order => 8 },
    "9" => { lower => 81, upper =>  90, order => 9 },
    "10" => { lower => 91, upper => 100, order => 10 },
};

# GPA categories
$score_cat{gpa} = {
    "3.9-4.0"  => { lower => 390, upper => 400, order => 1 },
    "3.8-3.89" => { lower => 380, upper => 389, order => 2 },
    "3.7-3.79" => { lower => 370, upper => 379, order => 3 },
    "3.6-3.69" => { lower => 360, upper => 369, order => 4 },
    "3.5-3.59" => { lower => 350, upper => 359, order => 5 },
    "3.4-3.49" => { lower => 340, upper => 349, order => 6 },
    "3.3-3.39" => { lower => 330, upper => 339, order => 7 },
    "3.2-3.29" => { lower => 320, upper => 329, order => 8 },
    "3.1-3.19" => { lower => 310, upper => 319, order => 9 },
    "3.0-3.09" => { lower => 300, upper => 309, order => 10 },
    "2.7-2.99" => { lower => 270, upper => 299, order => 11 },
    "2.5-2.69" => { lower => 250, upper => 269, order => 12 },
    "2.3-2.49" => { lower => 230, upper => 249, order => 13 },
    "2.0-2.29" => { lower => 200, upper => 229, order => 14 },
    "1.0-1.99" => { lower => 100, upper => 199, order => 15 },
    "< 1.0"    => { lower =>   1, upper =>  99, order => 16 },
};

### Subroutines

sub usage ($) {
    # Print usage statement and die with an error message, if
    # specified.
    my $choice;                 # individual score option

    print "Usage: $prog_name session year score_source\n\n";
    print "Choices for score_source are the following:\n";

    foreach $choice (sort keys %score_choices) {
        print $choice . " -- " . $score_choices{$choice} . "\n";
    }

    die "\nError: " . $_[0] . "!\n";
}  # end sub usage

sub interpret_parameters {
    # Interpret the parameters passed to us and refuse to run unless
    # they are sufficient and correct.

    # Perform tests to see if parameters are correct

    if ((! defined $ARGV[2]) || (defined $ARGV[3])) {
        &usage ("Incorrect number of parameters");
    }

    if ((length $ARGV[0] != 2) || (length $ARGV[1] != 4)) {
        &usage ("Parameters seem to be malformed");
    }

    if (! exists $score_choices{$ARGV[2]}) {
        &usage ("$ARGV[2] is not a valid option for score_source");
    }

    # Assign the rest of the parameters
    $sess = $ARGV[0];
    $yr = $ARGV[1];
    $score_type = $ARGV[2];

    # Make sure that we actually gave $score_q and %score_cat
    # meaningful data.
    if ((! defined $score_q{$score_type})
        || (! defined $score_cat{$score_type})) {
        die "The implementation of $score_type is not complete";
    }
}  # end sub interpret_parameters

sub create_temp_tables {
    # Make a bunch of temp tables that have all the data we need to
    # fetch later on.
    my ($query, $ret_val, $sth);

    # Open a connection to the database
    $dbh = DBI->connect('dbi:Informix:cars',
                        {PrintError => 1, RaiseError => 1})
        || die "Database connection not made: $DBI::errstr";

    # Compose SQL statement
    $query = $stu_progress_q . $score_q{$score_type};

    # Execute 1st SQL statement
    $sth = $dbh->prepare($stu_q);
    $sth->bind_param(1, $sess, SQL_VARCHAR);
    $sth->bind_param(2, $yr, SQL_INTEGER);
    $ret_val = $sth->execute();

    # Execute composite SQL statement
    $sth = $dbh->prepare($query);
    $ret_val = $sth->execute();

    # Check for errors
    if ($ret_val < 0) {
        die "Could not create temp tables";
    }
}  # end sub create_temp_tables

sub set_stu_progress {
    # Set stu_progress (a hash to admissions progress and an array of
    # students).
    my ($id, $progress, $ret_val, $row, $sth);

    # Compose SQL statement
    my $query = 'select id, progress from t_progress';

    # Execute SQL statement
    $sth = $dbh->prepare($query);
    $ret_val = $sth->execute();
    if ($ret_val < 0) {
        die "Could not fetch from t_progress";
    }

    # Interpret results
    while ($row = $sth->fetchrow_arrayref) {
        $id = $row->[0];
        $progress = $row->[1];
        $progress =~ s/^ +//g;  # strip leading whitespace
        $progress =~ s/ +$//g;  # strip trailing whitespace
        push @{ $stu_progress{$progress} }, $id;
    }

}  # end sub set_stu_progress

sub set_stu_score {
    # Set stu_score (a hash of students and score).
    my ($id, $ret_val, $row, $score, $sth);

    # Compose SQL statement
    my $query = 'select id, score from t_score';

    # Execute SQL statement
    $sth = $dbh->prepare($query);
    $ret_val = $sth->execute();
    if ($ret_val < 0) {
        die "Could not fetch from t_score";
    }

    # Interpret results
    while ($row = $sth->fetchrow_arrayref) {
        $id = $row->[0];
        $score = $row->[1];
        $stu_score{$id} = $score;
    }
}  # end sub set_stu_score

sub make_progress_score_cnt {
    # Make a hash of a hash of progress, score, and count of students,
    # assigning it to %progress_score_cnt.  This hash is not
    # guaranteed to have a count for every progress and cat.  You
    # should check using something like:
    #
    #     if (defined $progress_score_cnt{$progress}{$cat})
    #
    # before using a value.  Likewise, the N/A category is also not
    # guaranteed to be defined.
    my ($progress, $id, $cat);
    my $catp;               # did we categorize the current data item?

    foreach $progress (@progresses) {

        foreach $id (@{ $stu_progress{$progress} }) {
            $catp = 0;

            foreach $cat (keys %{ $score_cat{$score_type} }) {

                if ((defined $stu_score{$id}) &&
                    ($stu_score{$id} >=
                     $score_cat{$score_type}{$cat}{lower})
                    &&
                    ($stu_score{$id} <=
                     $score_cat{$score_type}{$cat}{upper}))
                {
                    # Increment the count for this progress and
                    # score_cat since it is between lower and upper
                    $progress_score_cnt{$progress}{$cat}++;
                    $catp = 1;  # we've counted it
                    last;
                }

            }  # next score category

            if (! $catp)      # we haven't counted it yet
            {
                $progress_score_cnt{$progress}{$na_cat{$score_type}}++;
            }

        }  # next id

    }  # next progress
}  # end sub set_progress_score_cnt

sub make_percentiles {
    # Determine the 25th and 75th percentiles for each student
    # category.
    my ($id, @list, $progress);

    # Loop through progresses
    foreach $progress (@progresses) {

        @list = ();

        # Generate list of usable ids
        foreach $id (@{ $stu_progress{$progress} }) {
            if (defined ${stu_score{$id}}) {
                push (@list, $id);
            }
        }
 
        # Sort list
        @list = (sort { $stu_score{$b} <=> $stu_score{$a} }
                 @list);

        # Find percentiles
        $percentiles{"25th"}{$progress} =
            $stu_score{$list[int($#list / 4)]};
        $percentiles{"75th"}{$progress} =
            $stu_score{$list[int($#list * 3 / 4)]};

        # Adjust to score type if necessary
        if (defined ${score_divisor{$score_type}}) {
            $percentiles{"25th"}{$progress} /=
                ${score_divisor{$score_type}};
            $percentiles{"75th"}{$progress} /=
                ${score_divisor{$score_type}};
        }

    }  # next progress
}  # end sub make_percentiles

sub make_progress_summary {
    # Find the average score and total count for each progress,
    # assigning results to %progress_summary.  The resulting hash of a
    # hash is guaranteed to have these two values for every progress.
    my ($count, $i, $id, $progress, $sum);

    # Assign total counts
    foreach $progress (@progresses) {
        $sum = 0;
        $count = 0;

        # Sum for each existing ID
        foreach $id (@{ $stu_progress{$progress} }) {
            if (defined ${stu_score{$id}}) {
                $sum += $stu_score{$id};
                $count++;
            }
        }  # next id

        # Adjust average, if necessary
        if (defined ${score_divisor{$score_type}}) {
            $sum /= ${score_divisor{$score_type}};
        }

        # Store values
        $progress_summary{$progress}{avg} = ($count > 0)
            ? $sum / $count : 0;
        $progress_summary{$progress}{count} = $count;
    }  # next progress
}  # end sub make_progress_summary

sub print_page_header {
    # Display the header for our document.  Called from
    # &display_results.
    my $date = `date`;

    print qq{<?xml version="1.0"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
  <head>
    <title>$prog_name -- $score_choices{$score_type}</title>
    <meta name="generator" content="admprofs.pl"/>
    <meta http-equiv="Content-Type"
          content="text/html; charset=utf-8"/>
    <style type="text/css">
      body {
        background-color: #dcdcdc;
        color: black;
      }
      h1, h2, h3 {
        text-align: center;
      }
      table {
        border-collapse: collapse;
      }
      .left {
        border-left: 2px solid black;
      }
      .right {
        border-right: 2px solid black;
      }
      .threeway {
        border-left: 2px solid black;
        border-right: 2px solid black;
        border-top: 2px solid black;
        border-bottom: 1px solid black;
      }
      td {
        border: 1px solid black;
        font-size: small;
        padding: 3px;
      }
      th {
        align: center;
        border: 1px solid black;
        padding: 3px;
      }
    </style>
  </head>
  <body>
<h1 id="top">$organization</h1>
<h2>$score_choices{$score_type} Profile / $student_type $sess $yr</h2>
<h3>$date</h3>

<table width="100%" style="background-color: white;">
  <col width="10%"/>
  <col width="11%"/><col width="11%"/>
  <col width="11%"/><col width="11%"/>
  <col width="12%"/>
  <col width="11%"/><col width="11%"/>
  <col width="12%"/>
  <tr style="background-color: #dcdcdc; border-width: 0;">
    <th style="border: 0;">&nbsp;</th>
    <th colspan="2" class="threeway">$progresses[0]</th>
    <th colspan="3" class="threeway">$progresses[1]</th>
    <th colspan="3" class="threeway">$progresses[2]</th>
  </tr>
  <tr>
    <th style="border: 0;">&nbsp;</th>
    <th class="left">Count</th><th class="right">%</th>
    <th class="left">Count</th><th>%</th>
    <th class="right">% Acc</th>
    <th class="left">Count</th><th>%</th>
    <th class="right">Yield</th>
  </tr>
};
}  # end sub print_page_header

sub print_cat_progress {
    # Print the first data table.  Called from &display_results.
    my ($cat, $progress);       # loop variables
    my $count;                  # count of students in cat x progress
    my $na_count;               # count of those in the N/A category
                                # for a particular progress
    my $percent;                # percentage of cat x progress over
                                # the total count of progress
    my $pc_change;              # percent change between categories
    my ($p_progress, $p_count); # values of the previous iteration

    # Each category
    foreach $cat (sort { $score_cat{$score_type}{$a}{order}
                         <=> $score_cat{$score_type}{$b}{order} }
                  keys %{ $score_cat{$score_type} }) {
        $p_progress = "";

        # Display line header
        print "  <tr>\n";
        print qq{    <td align="center"><strong>$cat</strong></td>\n};

        # Process for each category within the line
        foreach $progress (@progresses) {

            # Get count
            $count = (defined $progress_score_cnt{$progress}{$cat})
                ? $progress_score_cnt{$progress}{$cat} : 0;

            # Calculate percentage
            $percent = ($progress_summary{$progress}{count} > 0)
                ? sprintf "%.2f", $count /
                  $progress_summary{$progress}{count} * 100
                : "***";

            # Display information
            print qq{    <td class="left" align="right">$count</td>};

            # Display the percent change if we aren't on the first
            # iteration
            if ($p_progress) {
                print qq{<td align="right">}. $percent . "</td>\n";
                $pc_change = ($p_count > 0)
                    ? sprintf "%.2f", 100 - ($p_count - $count) * 100 / $p_count
                    : "***";
                print qq{    <td class="right" align="right">}
                . $pc_change . "</td>\n";
            } else {
                print qq{<td class="right" align="right">}
                . $percent . "</td>\n";
            }

            # Set the "previous" values
            $p_progress = $progress;
            $p_count = $count;

        }  # next progress

        # Display line footer
        print "  </tr>\n";
    }  # next cat
}  # end sub print_cat_progress

sub print_percentiles {
    # Print each of the percentiles that we have made.
    my ($progress, $p_progress, $ptile);

    foreach $ptile (sort keys %percentiles) {

        # Print line header
        print "  <tr>\n";
        print qq{    <td align="center"><strong>$ptile</strong></td>\n};
        $p_progress = "";

        foreach $progress (@progresses) {

            print q{    <td class="left" align="right">}
            . $percentiles{$ptile}{$progress} . "</td>";

            # Only print that extra column if we aren't on the first
            # iteration.
            if ($p_progress) {
                print qq{    <td>&nbsp;</td>\n};
                print qq{    <td class="right">&nbsp;</td>\n};
            } else {
                print qq{<td class="right">&nbsp;</td>\n};
            }

            $p_progress = $progress;
        }  # next progress

        # Print line footer
        print "  </tr>\n";

    }  # next ptile
}  # end sub print_percentiles

sub print_na_line {
    # Print the N/A line after every other category is displayed.
    # Called from &display_results.
    my ($na_count, $progress, $p_progress);

    # N/A line header
    print "  <tr>\n";
    print qq{    <td align="center"><strong>$na_cat{$score_type}}
        . "</strong></td>\n";
    $p_progress = "";

    # Loop through progresses
    foreach $progress (@progresses) {

        # Make sure na_count is defined
        $na_count =
            (defined
             $progress_score_cnt{$progress}{$na_cat{$score_type}})
            ? $progress_score_cnt{$progress}{$na_cat{$score_type}}
            : 0;

        print qq{    <td class="left" align="right">$na_count</td>};

        # Only print that extra column if we aren't on the first
        # iteration.
        if ($p_progress) {
            print qq{    <td>&nbsp;</td>\n};
            print qq{    <td class="right">&nbsp;</td>\n};
        } else {
            print qq{<td class="right">&nbsp;</td>\n};
        }

        $p_progress = $progress;
    }  # next progress

    # Print line footer
    print "  </tr>\n";
}  # end sub print_na_line

sub print_totals_line {
    # Print the totals line.  Called from &display_results.
    my ($count, $pc_change, $progress, $p_progress, $p_count);

    # Print line header
    print "  <tr>\n";
    print qq{    <td align="center"><strong>Totals</strong></td>\n};

    $p_progress = "";

    # Loop through progresses
    foreach $progress (@progresses) {

        $count = $progress_summary{$progress}{count};

        # Add N/A count
        $count += $progress_score_cnt{$progress}{$na_cat{$score_type}};

        print qq{    <td class="left" align="right">$count</td>\n};

        # Display the percent change if we aren't on the first
        # iteration
        if ($p_progress) {
            print qq{    <td>&nbsp;</td>\n};
            $pc_change = ($p_count > 0)
                ? sprintf "%.2f", 100 - ($p_count - $count) * 100 / $p_count
                : "***";
            print qq{    <td class="right" align="right">}
            . $pc_change . "</td>\n";
        } else {
            print qq{    <td class="right">&nbsp;</td>\n};
        }

        $p_progress = $progress;
        $p_count = $count;
    }  # next progress

    # Close this line and the table
    print "  </tr>\n</table>\n";
}  # end sub print_totals_line

sub print_progress_summary {
    # Print the progress summary table.  Called from &display_results.
    my ($avg, $progress, $total);

    # Print table header
    print qq{
<h2 style="margin-top: 40px;">Average $score_choices{$score_type}</h2>
<table width="100%" style="border-width: 0; border-style: invisible;">
  <col width="25%"/><col width="50%"/><col width="25%"/>
  <tr>
    <td style="border: 0;">&nbsp;</td>
    <td align="center" style="border: 0;">
<table style="background-color: white;">
};

    # Loop through progresses
    foreach $progress (@progresses) {
        $avg = $progress_summary{$progress}{avg};
        print qq{  <tr>\n    <td align="right" style="padding: 10px;">}
            . "<strong>" . $progress . "</strong></td>\n";
        print qq{    <td align="right" style="padding: 10px;">};
        printf "%.2f</td>\n", $avg;
        print "  </tr>\n";
    }

    # Close this table
    print q{</table>
    </td>
    <td style="border: 0;">&nbsp;</td>
  </tr>
</table>
};

}  # end sub print_progress_summary

sub print_page_footer {
    # Print footer

    print qq{
</body>
</html>
};

}  # end sub print_page_footer

sub display_results {
    # Display the results in the form of an XHTML 1.1 compliant
    # document to standard out.

    # Print page header
    &print_page_header;

    # Print the first data table
    &print_cat_progress;

    # Print the percentiles
    &print_percentiles;

    # Print the N/A line after every other category is displayed
    &print_na_line;

    # Print the totals line
    &print_totals_line;

    # Print the last data table
    &print_progress_summary;

    # Print page footer
    &print_page_footer;

}  # end sub display_results

### Main part of program

# Interpret the parameters that we have been given
&interpret_parameters;

# Make a temporary table in the database that has all of the data that
# we will need later
&create_temp_tables;

# For each progress, put the students that qualify for it into an
# array; the result: a hash of arrays
&set_stu_progress;

# Record the score of each student
&set_stu_score;

# Disconnect from the database; we don't need it anymore
$dbh->disconnect();

# Make a 3D hash of progress, score category, and count of students
&make_progress_score_cnt;

# Find the total count of students and average score for each
# progress.
&make_progress_summary;

# Compile percentiles
&make_percentiles;

# Display our findings in a lovely XHTML 1.1 compliant document
# to standard out.
&display_results;
